///////////////////////////////////////////////////////////////////////////////
// ViewGL.mm
// =========
// View & Controller components of OpenGL window
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2013-01-22
// UPDATED: 2021-09-09
//
//  Copyright (c) 2013 Song Ho Ahn. All rights reserved.
///////////////////////////////////////////////////////////////////////////////

#import "ViewGL.h"

#define FRAME_TIME 0.033f   // sec per frame

@implementation ViewGL
@synthesize timer;


///////////////////////////////////////////////////////////////////////////////
- (void)setModel: (ModelGL*)m
{
    model = m;
}



///////////////////////////////////////////////////////////////////////////////
// initialize member variables
- (void) awakeFromNib
{
    model = nil;
    timer = nil;
    printf("ViewGL: awakeFromNib\n");
}



///////////////////////////////////////////////////////////////////////////////
- (void) drawRect: (NSRect)dirtyRect
{
    if(!model)
        return;

    model->draw();
    [self.openGLContext flushBuffer];
}



///////////////////////////////////////////////////////////////////////////////
- (void)prepareOpenGL
{
    printf("ViewGL: prepareOpenGL\n");
    [super prepareOpenGL];
}



///////////////////////////////////////////////////////////////////////////////
- (void) reshape
{
    [super reshape];

    if(!model)
        return;

    NSRect rect = [self convertRectToBacking:[self bounds]];
    int width = (int)rect.size.width;
    int height = (int)rect.size.height;
    model->setWindowSize(width, height);
    [self setNeedsDisplay:YES];

    printf("OpenGL window is resized: %d x %d\n", width, height);
}



///////////////////////////////////////////////////////////////////////////////
// timer function for opengl rendering loop
- (void) startRenderingLoop
{
    /*@@ not work
    timer = [NSTimer scheduledTimerWithTimeInterval:FRAME_TIME
                                             target:self
                                           selector:@selector(loopFrames:)
                                           userInfo:nil
                                            repeats:YES];
    */
    //@@ the default run loop mode won't work mouseDown on a button.
    // add timer for common modes instead of default (system-defined) mode
    timer = [NSTimer timerWithTimeInterval:FRAME_TIME
                                    target:self
                                  selector:@selector(loopFrames:)
                                  userInfo:nil
                                   repeats:YES];
    [[NSRunLoop mainRunLoop] addTimer:timer
                              forMode:NSRunLoopCommonModes];
    //printf("timer started.\n");
}

- (void) stopRenderingLoop
{
    [timer invalidate];
    timer = nil;
    //printf("timer stopped.\n");
}

- (void) loopFrames: (NSTimer*)timer
{
    [self setNeedsDisplay:YES];
}



/*
///////////////////////////////////////////////////////////////////////////////
// override this function returns YES, so, the class can receive mouse event
- (BOOL) acceptsFirstMouse: (NSEvent*)event
{
    NSLog(@"here");
    return YES;
}
*/



///////////////////////////////////////////////////////////////////////////////
// handle mouse events
- (void) mouseDown: (NSEvent*)event
{
    if(!model)
        return;

    NSPoint pos = [NSEvent mouseLocation];
    model->setMousePosition((int)pos.x, (int)-pos.y);
}

- (void) mouseUp: (NSEvent*)event
{
    if(!model)
        return;

    NSPoint pos = [NSEvent mouseLocation];
    model->setMousePosition((int)pos.x, (int)-pos.y);
}

- (void) rightMouseDown: (NSEvent*)event
{
    if(!model)
        return;

    NSPoint pos = [NSEvent mouseLocation];
    model->setMousePosition((int)pos.x, (int)-pos.y);
}

- (void) rightMouseUp: (NSEvent*)event
{
    if(!model)
        return;

    NSPoint pos = [NSEvent mouseLocation];
    model->setMousePosition((int)pos.x, (int)-pos.y);
}

- (void) mouseDragged: (NSEvent*)event
{
    if(!model)
        return;

    NSPoint pos = [NSEvent mouseLocation];
    model->rotateCamera((int)pos.x, (int)-pos.y);
    [self setNeedsDisplay:YES];
}

- (void) rightMouseDragged:(NSEvent *)theEvent
{
    if(!model)
        return;

    NSPoint pos = [NSEvent mouseLocation];
    model->zoomCamera((int)-pos.y);
    [self setNeedsDisplay:YES];
}

- (void) scrollWheel: (NSEvent*)event
{
    if(!model)
        return;

    NSPoint pos = [NSEvent mouseLocation];
    int delta = (int)[event deltaY];
    model->setMousePosition((int)pos.x, (int)-pos.y);
    model->zoomCamera((int)-pos.y + delta);
    [self setNeedsDisplay:YES];
}

- (void)keyDown:(NSEvent*)event
{
    if(!model)
        return;

    //printf("KEY: %d\n", event.keyCode);
    if(event.keyCode == 123)
    {
        model->shiftCameraDelta(-1, 0);
        [self setNeedsDisplay:YES];
    }
    else if(event.keyCode == 124)
    {
        model->shiftCameraDelta(1, 0);
        [self setNeedsDisplay:YES];
    }
    else if(event.keyCode == 125)
    {
        model->shiftCameraDelta(0, 1);
        [self setNeedsDisplay:YES];
    }
    else if(event.keyCode == 126)
    {
        model->shiftCameraDelta(0, -1);
        [self setNeedsDisplay:YES];
    }
    else
    {
        [super keyDown:event];
    }
}

@end
